from typing import Final, Literal

VerificationMode = Literal["sampled", "full"]
SolanaCluster = Literal["solana-devnet", "solana-mainnet"]

VERIFICATION_MODES: Final[tuple[VerificationMode, ...]] = ("sampled", "full")
SOLANA_CLUSTERS: Final[tuple[SolanaCluster, ...]] = ("solana-devnet", "solana-mainnet")
