mod attestation;
mod models;
mod validation;

pub use attestation::{
    compute_attestation_signature, make_attestation, verify_attestation, Attestation,
};
pub use models::{ArtifactRecord, CaptureEnvelope, PublishEnvelope, TransformEnvelope};
pub use validation::{ValidationErrors, ValidationResult};

#[cfg(test)]
mod tests;
