from collections.abc import Sequence

from trazaeo_workflows.workflow_cli import main as workflow_main


def main(argv: Sequence[str] | None = None) -> None:
    workflow_main(argv)


if __name__ == "__main__":
    main()
