/// Tests that schema files are valid json.
#[test]
fn schema_files_are_valid_json() {
    let capture = include_str!("../schemas/capture.schema.json");
    let transform = include_str!("../schemas/transform.schema.json");
    let publish = include_str!("../schemas/publish.schema.json");
    let conformance = include_str!("../schemas/conformance_profiles.json");

    serde_json::from_str::<serde_json::Value>(capture).expect("capture schema valid json");
    serde_json::from_str::<serde_json::Value>(transform).expect("transform schema valid json");
    serde_json::from_str::<serde_json::Value>(publish).expect("publish schema valid json");
    serde_json::from_str::<serde_json::Value>(conformance).expect("conformance profile valid json");
}
